package djmil.cordacheckers.cordaclient;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.UUID;

import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

import djmil.cordacheckers.cordaclient.dao.GameState;
import djmil.cordacheckers.cordaclient.dao.GameState.Status;
import djmil.cordacheckers.cordaclient.dao.Stone;
import djmil.cordacheckers.user.HoldingIdentityResolver;

@SpringBootTest
public class GameBoardTests {
    @Autowired
    CordaClient cordaClient;

    @Autowired
    HoldingIdentityResolver holdingIdentityResolver;

    final String whitePlayerName = "alice";
    final String blackPlayerName = "bob";

    final static Stone WHITE_MAN  = new Stone(Stone.Color.WHITE, Stone.Type.MAN);
    final static Stone WHITE_KING = new Stone(Stone.Color.WHITE, Stone.Type.KING);
    final static Stone BLACK_MAN  = new Stone(Stone.Color.BLACK, Stone.Type.MAN);
    final static Stone BLACK_KING = new Stone(Stone.Color.BLACK, Stone.Type.KING);

    @Test
    void testSurrender() {
        final var hiWhite = holdingIdentityResolver.getByUsername(whitePlayerName);
        final var hiBlack = holdingIdentityResolver.getByUsername(blackPlayerName);

        final GameState game = cordaClient.gameProposalCreate(
            hiWhite, hiBlack, Stone.Color.BLACK, "GameBoard SURRENDER test");

        System.out.println("Game UUID " +game.uuid());

        final GameState acceptedGameBlackView = cordaClient.gameProposalAccept(
            hiBlack, game.uuid());

        assertThat(acceptedGameBlackView.opponentColor()).isEqualByComparingTo(Stone.Color.WHITE);
        assertThat(acceptedGameBlackView.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        // Black shall be able to surrender, even if it is not his turn
        final GameState surrendererGameView = cordaClient.gameBoardSurrender(
            hiBlack, game.uuid());

        assertThat(surrendererGameView.opponentName()).isEqualToIgnoringCase(whitePlayerName);
        assertThat(surrendererGameView.opponentColor()).isEqualByComparingTo(Stone.Color.WHITE);
        assertThat(surrendererGameView.status()).isEqualByComparingTo(Status.GAME_RESULT_YOU_LOOSE);

        assertThatThrownBy(() -> { // White can not surrender, since Black already did
            cordaClient.gameBoardSurrender(
                hiWhite, game.uuid());
        });

        final GameState winnerGameView = cordaClient.gameStateGet(
            hiWhite, game.uuid());

        assertThat(winnerGameView.opponentName()).isEqualToIgnoringCase(blackPlayerName);
        assertThat(winnerGameView.opponentColor()).isEqualByComparingTo(Stone.Color.BLACK);
        assertThat(winnerGameView.status()).isEqualByComparingTo(Status.GAME_RESULT_YOU_WON);
    }

    @Test
    void testMove() {
        final var hiWhite = holdingIdentityResolver.getByUsername(whitePlayerName);
        final var hiBlack = holdingIdentityResolver.getByUsername(blackPlayerName);

        final GameState game = cordaClient.gameProposalCreate(
            hiWhite, hiBlack, Stone.Color.BLACK, "GameBoard MOVE test");
        System.out.println("Game UUID " +game.uuid());

        final var m0 = cordaClient.gameProposalAccept(hiBlack, game.uuid());
        assertThat(m0.previousMove()).isNull();
        assertThat(m0.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiBlack, game.uuid(), move(12, 16), 
                "Black can not move, since it is opponent's turn");
        });

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiBlack, game.uuid(), move(24, 20), 
                "Black can not move opponent's stone as well");
        });

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiWhite, game.uuid(), move(9, 13),
            "White can not move opponent's stone");
        });

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiWhite, game.uuid(), move(17, 14),
            "Trying to move an empty tile");
        });

        final var m1 = cordaClient.gameBoardMove(hiWhite, game.uuid(), move(22, 18), null);
        assertThat(m0.board().get(22)).isEqualTo(WHITE_MAN);
        assertThat(m0.board().get(18)).isNull();
        assertThat(m1.board().get(22)).isNull();
        assertThat(m1.board().get(18)).isEqualTo(WHITE_MAN);
        assertThat(m1.moveNumber() == 1);
        assertThat(m1.previousMove().get(0) == 22);
        assertThat(m1.previousMove().get(1) == 18);
        assertThat(m1.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiBlack, game.uuid(), move(12, 13), 
            "Prohibited move shall be rejected");
        });

        final var m2 = cordaClient.gameBoardMove(hiBlack, game.uuid(), move(11, 15), null);
        assertThat(m1.board().get(11)).isEqualTo(BLACK_MAN);
        assertThat(m1.board().get(15)).isNull();
        assertThat(m2.board().get(11)).isNull();
        assertThat(m2.board().get(15)).isEqualTo(BLACK_MAN);
        assertThat(m2.moveNumber() == 2);
        assertThat(m2.previousMove().get(0) == 11);
        assertThat(m2.previousMove().get(1) == 15);
        assertThat(m2.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiWhite, game.uuid(), move(24, 19), 
            "An attempt to ignore mandatory capture");
        });

        final var m3 = cordaClient.gameBoardMove(hiWhite, game.uuid(), move(18, 11), null);
        assertThat(m2.board().get(18)).isEqualTo(WHITE_MAN);
        assertThat(m2.board().get(11)).isNull();
        assertThat(m2.board().get(15)).isEqualTo(BLACK_MAN);
        assertThat(m3.board().get(11)).isEqualTo(WHITE_MAN);
        assertThat(m3.board().get(18)).isNull();
        assertThat(m3.board().get(15)).isNull();
        assertThat(m3.moveNumber() == 3);
        assertThat(m3.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        final var m4 = cordaClient.gameBoardMove(hiBlack, game.uuid(), move(8, 15), 
            "Choose one of two possible mandatory jumps");
        assertThat(m4.moveNumber() == 4);
        assertThat(m4.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        final var m5 = cordaClient.gameBoardMove(hiWhite, game.uuid(), move(26, 22), null);
        assertThat(m5.moveNumber() == 5);
        assertThat(m5.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        final var m6 = cordaClient.gameBoardMove(hiBlack, game.uuid(), move(4, 8),  null);
        assertThat(m6.moveNumber() == 6);
        assertThat(m6.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        final var m7 = cordaClient.gameBoardMove(hiWhite, game.uuid(), move(23, 18),
            "A move that ends in mandatory capture on a next turn for me");
        assertThat(m7.moveNumber() == 7);
        assertThat(m7.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        final var m8 = cordaClient.gameBoardMove(hiBlack, game.uuid(), move(9, 14),
            "Moard mandatory capture on a next turn for opponent");
        assertThat(m8.moveNumber() == 8);
        assertThat(m8.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);

        final var m9 = cordaClient.gameBoardMove(hiWhite, game.uuid(), move(18, 11),
        "First part of a mandatory jump");
        assertThat(m9.moveNumber() == 9);
        assertThat(m9.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_YOU);

        assertThatThrownBy(() -> { 
            cordaClient.gameBoardMove(hiWhite, game.uuid(), move(22, 17), 
            "An attempt to ignore mandatory capture");
        });
    
        final var m10 = cordaClient.gameBoardMove(hiWhite, game.uuid(), move(11, 4),
        "Second part of a mandatory jump into a king row");
        assertThat(m10.board().get(4)).isEqualTo(WHITE_KING);
        assertThat(m10.moveNumber() == 9);
        assertThat(m10.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);
    }

    ArrayList<Integer> move(int from, int to) {
        return new ArrayList<Integer>(Arrays.asList(from, to));
    }
}
