package djmil.cordacheckers.cordaclient;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;

import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

import djmil.cordacheckers.cordaclient.dao.GameState;
import djmil.cordacheckers.cordaclient.dao.GameState.Status;
import djmil.cordacheckers.cordaclient.dao.Stone;
import djmil.cordacheckers.user.HoldingIdentityResolver;

@SpringBootTest
public class GameProposalTests {
    @Autowired
    CordaClient cordaClient;

    @Autowired
    HoldingIdentityResolver holdingIdentityResolver;

    final String issuer = "alice";
    final String acquier = "bob";
    final Stone.Color acquierColor = Stone.Color.WHITE;

    @Test
    void testCreate() {
        final var hiIssuer  = holdingIdentityResolver.getByUsername(issuer);
        final var hiAcquier = holdingIdentityResolver.getByUsername(acquier);
        final String message = "GameProposal CREATE test";

        final GameState issuerGameView = cordaClient.gameProposalCreate(
            hiIssuer,
            hiAcquier,
            acquierColor,
            message);

        System.out.println("Game UUID " +issuerGameView.uuid());

        /*
         * Issuers perspective on a newly created GameProposal
         */
        assertThat(issuerGameView.status()).isEqualByComparingTo(Status.GAME_PROPOSAL_WAIT_FOR_OPPONENT);
        assertThat(issuerGameView.opponentName()).isEqualToIgnoringCase(acquier);
        assertThat(issuerGameView.opponentColor()).isEqualByComparingTo(acquierColor);
        assertThat(issuerGameView.message()).isEqualToIgnoringCase(message);
        assertThat(issuerGameView.uuid()).isNotNull();
      
        /*
         * Acquier shall be able to find newly created GameProposal
         */
        GameState acquierGameView = assertDoesNotThrow( () -> {
            return cordaClient.gameStateList(hiAcquier)
                .stream()
                .filter(gameView -> gameView.uuid().equals(issuerGameView.uuid()))
                .reduce((a, b) -> {throw new IllegalStateException("Duplicate");})
                .get();
        });

        assertThat(acquierGameView.status()).isEqualByComparingTo(Status.GAME_PROPOSAL_WAIT_FOR_YOU);
        assertThat(acquierGameView.opponentName()).isEqualToIgnoringCase(issuer);
        assertThat(acquierGameView.opponentColor()).isEqualByComparingTo(acquierColor.opposite());
        assertThat(acquierGameView.message()).isEqualToIgnoringCase(message);
        assertThat(acquierGameView.uuid()).isEqualByComparingTo(issuerGameView.uuid());
    }

    @Test
    void testReject() {
        final var hiIssuer  = holdingIdentityResolver.getByUsername(issuer);
        final var hiAcquier = holdingIdentityResolver.getByUsername(acquier);
        final String message = "GameProposal REJECT test";

        final GameState game = cordaClient.gameProposalCreate(
            hiIssuer, hiAcquier, acquierColor, message);

        System.out.println("Game UUID " +game.uuid());

        assertThatThrownBy(() -> {  // Issuer can not reject
            cordaClient.gameProposalReject(
                hiIssuer, game.uuid());
        });

        final GameState pendingGameView = cordaClient.gameStateGet(
            hiAcquier, game.uuid());
        assertThat(pendingGameView.status()).isEqualByComparingTo(Status.GAME_PROPOSAL_WAIT_FOR_YOU);
        assertThat(pendingGameView.uuid()).isEqualByComparingTo(game.uuid());

        final GameState rejectedGameView = cordaClient.gameProposalReject(
            hiAcquier, game.uuid());
        assertThat(rejectedGameView.status()).isEqualByComparingTo(Status.GAME_PROPOSAL_REJECTED);
        assertThat(rejectedGameView.uuid()).isEqualByComparingTo(game.uuid());

        assertThatThrownBy(() -> {  // Issuer has no pending GameProposal
            cordaClient.gameStateGet(
                hiIssuer, game.uuid());
        });
    }

    @Test
    void testCancel() {
        final var hiIssuer  = holdingIdentityResolver.getByUsername(issuer);
        final var hiAcquier = holdingIdentityResolver.getByUsername(acquier);
        final String message = "GameProposal CANCEL test";

        final GameState game = cordaClient.gameProposalCreate(
            hiIssuer, hiAcquier, acquierColor, message);

        System.out.println("Game UUID " +game.uuid());

        assertThatThrownBy(() -> {  // Acquier can not cancel
            cordaClient.gameProposalCancel(
                hiAcquier, game.uuid());
        });

        final GameState pendingGameView = cordaClient.gameStateGet(
            hiIssuer, game.uuid());
        assertThat(pendingGameView.status()).isEqualByComparingTo(Status.GAME_PROPOSAL_WAIT_FOR_OPPONENT);
        assertThat(pendingGameView.uuid()).isEqualByComparingTo(game.uuid());

        final GameState canceledGameView = cordaClient.gameProposalCancel(
            hiIssuer, game.uuid());
        assertThat(canceledGameView.status()).isEqualByComparingTo(Status.GAME_PROPOSAL_CANCELED);
        assertThat(canceledGameView.uuid()).isEqualByComparingTo(game.uuid());

        assertThatThrownBy(() -> {  // Acquier has no pending GameProposal
            cordaClient.gameStateGet(
            hiAcquier, game.uuid());
        });
    }

    @Test
    void testAccept() {
        final var hiIssuer  = holdingIdentityResolver.getByUsername(issuer);
        final var hiAcquier = holdingIdentityResolver.getByUsername(acquier);
        final String message = "GameProposal ACCEPT test";

        final GameState game = cordaClient.gameProposalCreate(
            hiIssuer, hiAcquier, acquierColor, message);

        System.out.println("Game UUID " +game.uuid());

        assertThatThrownBy(() -> { // Issuer can not accept
            cordaClient.gameProposalAccept(
                hiIssuer, game.uuid());
        });

        final GameState acceptedGameAcquierView = cordaClient.gameProposalAccept(
            hiAcquier, game.uuid());

        assertThat(acceptedGameAcquierView.opponentName()).isEqualToIgnoringCase(issuer);
        assertThat(acceptedGameAcquierView.opponentColor()).isEqualByComparingTo(acquierColor.opposite());
        assertThat(acceptedGameAcquierView.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_YOU);
        assertThat(acceptedGameAcquierView.uuid()).isEqualByComparingTo(game.uuid());

        assertThatThrownBy(() -> {   // same GameProposal can not be accepted twice 
            cordaClient.gameProposalAccept(
                hiIssuer, game.uuid());
        });

        final GameState acceptedGameIssuerView = cordaClient.gameStateGet(
            hiIssuer, game.uuid());

        assertThat(acceptedGameIssuerView.opponentName()).isEqualToIgnoringCase(acquier);
        assertThat(acceptedGameIssuerView.opponentColor()).isEqualByComparingTo(acquierColor);
        assertThat(acceptedGameIssuerView.status()).isEqualByComparingTo(Status.GAME_BOARD_WAIT_FOR_OPPONENT);
        assertThat(acceptedGameIssuerView.uuid()).isEqualByComparingTo(game.uuid());
    }

}
