package djmil.cordacheckers.contracts;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleInputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;

import java.util.List;

import djmil.cordacheckers.checkers.Move;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.GameResultState;
import djmil.cordacheckers.states.GameState;
import net.corda.v5.base.annotations.ConstructorForDeserialization;
import net.corda.v5.base.annotations.CordaSerializable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

public class GameCommand implements Command {
    @CordaSerializable
    public static enum Action {
        GAME_PROPOSAL_CREATE,
        GAME_PROPOSAL_CANCEL,
        GAME_PROPOSAL_ACCEPT,
        GAME_PROPOSAL_REJECT,

        GAME_BOARD_MOVE,
        GAME_BOARD_SURRENDER,
        GAME_BOARD_VICTORY;
    }

    private final Action action;
    private final List<Integer> move;

    public static class ActionException extends RuntimeException {
        public ActionException() {
            super("Bad GameCommand.Action value");
        }
    }

    // Serialisation service requires a default constructor
    public GameCommand() {
        this.action = null;
        this.move = null;
    }

    public GameCommand(Action action) {
        if (action == Action.GAME_BOARD_MOVE)
            throw new ActionException();

        this.action = action;
        this.move = null;
    }

    public GameCommand(List<Integer> move) {
        this.action = Action.GAME_BOARD_MOVE;
        this.move = move;
    }

    @ConstructorForDeserialization
    public GameCommand(Action action, List<Integer> move) {
        this.action = action;
        this.move = move;
    }

    public Action getAction() {
        return action;
    }

    public List<Integer> getMove() {
        return move;
    }

    public MemberX500Name getCounterparty(StateAndRef<GameState> gameStateSar) {
        return getCounterparty(gameStateSar.getState().getContractState());
    }
    
    public MemberX500Name getCounterparty(GameState gameState) {
        switch (this.action) {
            case GAME_PROPOSAL_CREATE:
            case GAME_PROPOSAL_CANCEL:
                if (gameState instanceof GameProposalState)
                    return ((GameProposalState)gameState).getAcquierName();
                break;

            case GAME_PROPOSAL_ACCEPT:
            case GAME_PROPOSAL_REJECT:
                if (gameState instanceof GameProposalState)
                    return ((GameProposalState)gameState).getIssuerName();
                break;

            case GAME_BOARD_MOVE:
                if (gameState instanceof GameBoardState)
                    return ((GameBoardState)gameState).getIdelPlayerName();
                break;
                
            case GAME_BOARD_SURRENDER:
                if (gameState instanceof GameResultState)
                    return ((GameResultState)gameState).getWinnerName();
                break;
            
            case GAME_BOARD_VICTORY:
                if (gameState instanceof GameResultState)
                    return ((GameResultState)gameState).getLooserName();
                break;
            
            default:
                throw new ActionException();
        }

        throw new RuntimeException(action +": unexpected GameState type " +gameState.getClass());
    }

    public void validateGameProposalCreate(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().isEmpty(), CREATE_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, CREATE_OUTPUT_STATE);

        final GameProposalState gameProposal = getSingleOutputState(trx, GameProposalState.class);

        /*
         * Major command logick check 
         */
        requireThat(gameProposal.getIssuerName().compareTo(gameProposal.getBlackPlayer()) == 0 ||
                    gameProposal.getIssuerName().compareTo(gameProposal.getWhitePlayer()) == 0,
            "GameProposal.Issuer must be either Black or White player");
    }

    public void validateGameProposalAccept(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, ACCEPT_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, ACCEPT_OUTPUT_STATE);

        final GameProposalState inGameProposal = getSingleInputState (trx, GameProposalState.class);
        final GameBoardState outGameBoard      = getSingleOutputState(trx, GameBoardState.class);

        requireThat(outGameBoard.getWhitePlayer().compareTo(inGameProposal.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getBlackPlayer().compareTo(inGameProposal.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getParticipants().containsAll(inGameProposal.getParticipants()), IN_OUT_PARTICIPANTS);

        /*
         * Major command logick check 
         */
        requireThat(outGameBoard.getBoard().equals(GameBoardContract.initialBoard), "Bad GameBoard initial state");
    }

    public void validateGameProposalReject(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, REJECT_INPUT_STATE);
        requireThat(trx.getOutputContractStates().isEmpty(), REJECT_OUTPUT_STATE);

        getSingleInputState(trx, GameProposalState.class);
    }

    public void validateGameProposalCancel(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, CANCEL_INPUT_STATE);
        requireThat(trx.getOutputContractStates().isEmpty(), CANCEL_OUTPUT_STATE);
        
        getSingleInputState(trx, GameProposalState.class);
    }

    public void validateGameBoardMove(UtxoLedgerTransaction trx, List<Integer> move) {
        requireThat(trx.getInputContractStates().size()  == 1, MOVE_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, MOVE_OUTPUT_STATE);

        final var inGameBoard  = getSingleInputState (trx, GameBoardState.class);
        final var outGameBoard = getSingleOutputState(trx, GameBoardState.class);

        requireThat(outGameBoard.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);

        /*
         * Major command logick check 
         */
        final var newGameBoard = new GameBoardState(inGameBoard, move, outGameBoard.getMessage());
        requireThat(outGameBoard.equals(newGameBoard), "Unexpected output state");
    }

    public void validateGameBoardSurrender(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, SURRENDER_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, SURRENDER_OUTPUT_STATE);

        final var inGameBoard   = getSingleInputState (trx, GameBoardState.class);
        final var outGameResult = getSingleOutputState(trx, GameResultState.class);

        requireThat(outGameResult.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);

        /*
         * Major command logick check 
         */
        requireThat(outGameResult.getLooserName().compareTo(outGameResult.getBlackPlayer()) == 0 ||
                    outGameResult.getLooserName().compareTo(outGameResult.getWhitePlayer()) == 0,
            "Surenderer must be either Black or White player");
    }

    public void validateGameBoardVictory(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, VICTORY_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, VICTORY_OUTPUT_STATE);

        final var inGameBoard   = getSingleInputState (trx, GameBoardState.class);
        final var outGameResult = getSingleOutputState(trx, GameResultState.class);

        requireThat(outGameResult.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);

        /*
         * Major command logick check 
         */
        final var possibleMoves = Move.getPossibleMoves(inGameBoard.getBoard(), inGameBoard.getActiveColor());
        requireThat(possibleMoves.isEmpty(), "Victory condition violation");

        requireThat(outGameResult.getWinnerName().compareTo(inGameBoard.getIdelPlayerName()) == 0, "Bad winner name");
    }

    public static void requireThat(boolean asserted, String errorMessage) {
        if (!asserted) {
            throw new IllegalStateException("Failed requirement: " + errorMessage);
        }
    }

    static final String REQUIRE_SINGLE_COMMAND = "Require a single command";
    static final String IN_OUT_PARTICIPANTS = "Output participants should include all input participants";

    static final String CREATE_INPUT_STATE  = "Create command should have no input states";
    static final String CREATE_OUTPUT_STATE = "Create command should output exactly one GameProposal state";

    static final String REJECT_INPUT_STATE  = "Reject command should have exactly one GameProposal input state";
    static final String REJECT_OUTPUT_STATE = "Reject command should have no output states";

    static final String CANCEL_INPUT_STATE  = "Cancel command should have exactly one GameProposal input state";
    static final String CANCEL_OUTPUT_STATE = "Cancel command should have no output states";

    static final String ACCEPT_INPUT_STATE  = "Accept command should have exactly one GameProposal input state";
    static final String ACCEPT_OUTPUT_STATE = "Accept command should have exactly one GameBoard output state";

    static final String SURRENDER_INPUT_STATE = "SURRENDER command should have exactly one GameBoardState input state";
    static final String SURRENDER_OUTPUT_STATE = "SURRENDER command should have exactly one GameResultState output state";

    static final String MOVE_INPUT_STATE = "MOVE command should have exactly one GameBoardState input state";
    static final String MOVE_OUTPUT_STATE = "MOVE command should have exactly one GameBoardState output state";
    static final String MOVE_OUT_BOARD = "MOVE command: move checkers rules violation";
    static final String MOVE_OUT_COLOR = "MOVE command: moveColor checkers rules violation";

    static final String VICTORY_INPUT_STATE  = "VICTORY command should have exactly one GameBoardState input state";
    static final String VICTORY_OUTPUT_STATE = "VICTORY command should have exactly one GameResultState output state";
}
