package djmil.cordacheckers.gamestate;


import djmil.cordacheckers.contracts.GameCommand;
import djmil.cordacheckers.contracts.GameInfo;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.GameResultState;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.SubFlow;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

/*
 * A view on a ledger transaction that has GameState involved in it, from perspective of a concrete player
 */ 
public class ViewBuilder implements SubFlow<View> {
    private final SecureHash gameStateUtxoTrxId;

    public ViewBuilder (SecureHash gameStateUtxoTrxId) {
        this.gameStateUtxoTrxId = gameStateUtxoTrxId;
    }

    @CordaInject
    public UtxoLedgerService utxoLedgerService;

    @CordaInject
    public MemberLookup memberLookup;

    @Override
    @Suspendable
    public View call() {
        final UtxoLedgerTransaction gameStateUtxo = this.utxoLedgerService
            .findLedgerTransaction(gameStateUtxoTrxId);

        return buildGameStateView(gameStateUtxo);
    }

    @Suspendable
    View buildGameStateView(UtxoLedgerTransaction gameStateUtxo) {
        final GameInfo game = new GameInfo(gameStateUtxo);
        
        final MemberX500Name myName = memberLookup.myInfo().getName();
        final View.Status viewStatus = action2status(game, myName);

        if (game.state instanceof GameProposalState)
            return new View(viewStatus, myName, (GameProposalState) game.state);

        if (game.state instanceof GameBoardState)
            return new View(viewStatus, myName, (GameBoardState) game.state, game.command.getMove());

        if (game.state instanceof GameResultState)
            return new View(viewStatus, myName, (GameResultState) game.state);

        throw new RuntimeException("Unexpected GameState");
    }

    @Suspendable
    View.Status action2status(GameInfo game, MemberX500Name myName) {

        switch (game.command.getAction()) {
            case GAME_PROPOSAL_CREATE:
                if (game.isMyAction(myName))
                    return View.Status.GAME_PROPOSAL_WAIT_FOR_YOU;
                else
                    return View.Status.GAME_PROPOSAL_WAIT_FOR_OPPONENT;
                    
            case GAME_PROPOSAL_REJECT:
                return View.Status.GAME_PROPOSAL_REJECTED;

            case GAME_PROPOSAL_CANCEL:
                return View.Status.GAME_PROPOSAL_CANCELED;
            
            case GAME_PROPOSAL_ACCEPT:
            case GAME_BOARD_MOVE:
                if (game.isMyAction(myName))
                    return View.Status.GAME_BOARD_WAIT_FOR_YOU;
                else
                    return View.Status.GAME_BOARD_WAIT_FOR_OPPONENT;

            case GAME_BOARD_SURRENDER:
                if (game.issuer.compareTo(myName) == 0)
                    return View.Status.GAME_RESULT_YOU_LOOSE;
                else
                    return View.Status.GAME_RESULT_YOU_WON; 
            
            case GAME_BOARD_VICTORY:
                if (game.issuer.compareTo(myName) == 0)
                    return View.Status.GAME_RESULT_YOU_WON;
                else
                    return View.Status.GAME_RESULT_YOU_LOOSE; 
        }

        throw new GameCommand.ActionException();
    }
}
